@@ -294,6 +294,36 @@ macdrv_window macdrv_get_cocoa_window(HWND hwnd, BOOL require_on_screen)
 }
 
 
+/***********************************************************************
+ *              macdrv_get_cocoa_view
+ *
+ * Return the Cocoa view associated with a window
+ */
+macdrv_view macdrv_get_cocoa_view(HWND hwnd)
+{
+    struct macdrv_win_data *data = get_win_data(hwnd);
+    macdrv_view ret = data ? data->cocoa_view : NULL;
+
+    release_win_data(data);
+    return ret;
+}
+
+
+/***********************************************************************
+ *              macdrv_get_client_cocoa_view
+ *
+ * Return the Cocoa view associated with a window's client area
+ */
+macdrv_view macdrv_get_client_cocoa_view(HWND hwnd)
+{
+    struct macdrv_win_data *data = get_win_data(hwnd);
+    macdrv_view ret = data ? data->client_cocoa_view : NULL;
+
+    release_win_data(data);
+    return ret;
+}
+
+
 /***********************************************************************
  *              set_cocoa_window_properties
  *
@@ -603,6 +633,27 @@ static void sync_window_min_max_info(HWND hwnd)
 }
 
 
+/**********************************************************************
+ *              create_client_cocoa_view
+ *
+ * Create the Cocoa view for a window's client area
+ */
+static void create_client_cocoa_view(struct macdrv_win_data *data)
+{
+    RECT rect = data->client_rect;
+    OffsetRect(&rect, -data->whole_rect.left, -data->whole_rect.top);
+
+    if (data->client_cocoa_view)
+        macdrv_set_view_frame(data->client_cocoa_view, cgrect_from_rect(rect));
+    else
+    {
+        data->client_cocoa_view = macdrv_create_view(cgrect_from_rect(rect));
+        macdrv_set_view_hidden(data->client_cocoa_view, FALSE);
+    }
+    macdrv_set_view_superview(data->client_cocoa_view, data->cocoa_view, data->cocoa_window, NULL, NULL);
+}
+
+
 /**********************************************************************
  *              create_cocoa_window
  *
@@ -646,6 +697,7 @@ static void create_cocoa_window(struct macdrv_win_data *data)
 
     data->cocoa_window = macdrv_create_cocoa_window(&wf, frame, data->hwnd, thread_data->queue);
     if (!data->cocoa_window) goto done;
+    create_client_cocoa_view(data);
 
     set_cocoa_window_properties(data);
 
@@ -687,6 +739,52 @@ static void destroy_cocoa_window(struct macdrv_win_data *data)
 }
 
 
+/**********************************************************************
+ *              create_cocoa_view
+ *
+ * Create the Cocoa view for a given Windows child window
+ */
+static void create_cocoa_view(struct macdrv_win_data *data)
+{
+    BOOL equal = !memcmp(&data->window_rect, &data->client_rect, sizeof(data->whole_rect));
+    CGRect frame = cgrect_from_rect(data->window_rect);
+
+    data->shaped = FALSE;
+    data->whole_rect = data->window_rect;
+
+    TRACE("creating %p window %s whole %s client %s\n", data->hwnd, wine_dbgstr_rect(&data->window_rect),
+          wine_dbgstr_rect(&data->whole_rect), wine_dbgstr_rect(&data->client_rect));
+
+    if (!equal)
+        data->cocoa_view = macdrv_create_view(frame);
+    create_client_cocoa_view(data);
+    if (equal)
+    {
+        data->cocoa_view = data->client_cocoa_view;
+        macdrv_set_view_hidden(data->cocoa_view, TRUE);
+        macdrv_set_view_frame(data->cocoa_view, frame);
+    }
+}
+
+
+/**********************************************************************
+ *              destroy_cocoa_view
+ *
+ * Destroy the Cocoa view for a given window.
+ */
+static void destroy_cocoa_view(struct macdrv_win_data *data)
+{
+    if (!data->cocoa_view) return;
+
+    TRACE("win %p Cocoa view %p\n", data->hwnd, data->cocoa_view);
+
+    if (data->cocoa_view != data->client_cocoa_view)
+        macdrv_dispose_view(data->cocoa_view);
+    data->cocoa_view = NULL;
+    data->on_screen = FALSE;
+}
+
+
 /***********************************************************************
  *              macdrv_create_win_data
  *
@@ -721,6 +819,13 @@ static struct macdrv_win_data *macdrv_create_win_data(HWND hwnd, const RECT *win
                hwnd, data->cocoa_window, wine_dbgstr_rect(&data->window_rect),
                wine_dbgstr_rect(&data->whole_rect), wine_dbgstr_rect(&data->client_rect));
     }
+    else
+    {
+        create_cocoa_view(data);
+        TRACE("win %p/%p window %s whole %s client %s\n",
+               hwnd, data->cocoa_view, wine_dbgstr_rect(&data->window_rect),
+               wine_dbgstr_rect(&data->whole_rect), wine_dbgstr_rect(&data->client_rect));
+    }
 
     return data;
 }
@@ -731,40 +836,50 @@ static struct macdrv_win_data *macdrv_create_win_data(HWND hwnd, const RECT *win
  */
 static void show_window(struct macdrv_win_data *data)
 {
-    HWND prev = NULL;
-    HWND next = NULL;
-    macdrv_window prev_window = NULL;
-    macdrv_window next_window = NULL;
-    BOOL activate = FALSE;
-    HWND hwndFocus;
-
-    /* find window that this one must be after */
-    prev = GetWindow(data->hwnd, GW_HWNDPREV);
-    while (prev && !((GetWindowLongW(prev, GWL_STYLE) & (WS_VISIBLE | WS_MINIMIZE)) == WS_VISIBLE &&
-                     (prev_window = macdrv_get_cocoa_window(prev, TRUE))))
-        prev = GetWindow(prev, GW_HWNDPREV);
-    if (!prev_window)
+    if (data->cocoa_window)
     {
-        /* find window that this one must be before */
-        next = GetWindow(data->hwnd, GW_HWNDNEXT);
-        while (next && !((GetWindowLongW(next, GWL_STYLE) & (WS_VISIBLE | WS_MINIMIZE)) == WS_VISIBLE &&
-                         (next_window = macdrv_get_cocoa_window(next, TRUE))))
-            next = GetWindow(next, GW_HWNDNEXT);
+        HWND prev = NULL;
+        HWND next = NULL;
+        macdrv_window prev_window = NULL;
+        macdrv_window next_window = NULL;
+        BOOL activate = FALSE;
+        HWND hwndFocus;
+
+        /* find window that this one must be after */
+        prev = GetWindow(data->hwnd, GW_HWNDPREV);
+        while (prev && !((GetWindowLongW(prev, GWL_STYLE) & (WS_VISIBLE | WS_MINIMIZE)) == WS_VISIBLE &&
+                         (prev_window = macdrv_get_cocoa_window(prev, TRUE))))
+            prev = GetWindow(prev, GW_HWNDPREV);
+        if (!prev_window)
+        {
+            /* find window that this one must be before */
+            next = GetWindow(data->hwnd, GW_HWNDNEXT);
+            while (next && !((GetWindowLongW(next, GWL_STYLE) & (WS_VISIBLE | WS_MINIMIZE)) == WS_VISIBLE &&
+                             (next_window = macdrv_get_cocoa_window(next, TRUE))))
+                next = GetWindow(next, GW_HWNDNEXT);
+        }
+
+        TRACE("win %p/%p below %p/%p above %p/%p\n",
+              data->hwnd, data->cocoa_window, prev, prev_window, next, next_window);
+
+        if (!prev_window)
+            activate = activate_on_focus_time && (GetTickCount() - activate_on_focus_time < 2000);
+        macdrv_order_cocoa_window(data->cocoa_window, prev_window, next_window, activate);
+        data->on_screen = TRUE;
+
+        hwndFocus = GetFocus();
+        if (hwndFocus && (data->hwnd == hwndFocus || IsChild(data->hwnd, hwndFocus)))
+            macdrv_SetFocus(hwndFocus);
+        if (activate)
+            activate_on_focus_time = 0;
     }
+    else
+    {
+        TRACE("win %p/%p showing view\n", data->hwnd, data->cocoa_view);
 
-    TRACE("win %p/%p below %p/%p above %p/%p\n",
-          data->hwnd, data->cocoa_window, prev, prev_window, next, next_window);
-
-    if (!prev_window)
-        activate = activate_on_focus_time && (GetTickCount() - activate_on_focus_time < 2000);
-    macdrv_order_cocoa_window(data->cocoa_window, prev_window, next_window, activate);
-    data->on_screen = TRUE;
-
-    hwndFocus = GetFocus();
-    if (hwndFocus && (data->hwnd == hwndFocus || IsChild(data->hwnd, hwndFocus)))
-        macdrv_SetFocus(hwndFocus);
-    if (activate)
-        activate_on_focus_time = 0;
+        macdrv_set_view_hidden(data->cocoa_view, FALSE);
+        data->on_screen = TRUE;
+    }
 }
 
 
@@ -775,11 +890,58 @@ static void hide_window(struct macdrv_win_data *data)
 {
     TRACE("win %p/%p\n", data->hwnd, data->cocoa_window);
 
-    macdrv_hide_cocoa_window(data->cocoa_window);
+    if (data->cocoa_window)
+        macdrv_hide_cocoa_window(data->cocoa_window);
+    else
+        macdrv_set_view_hidden(data->cocoa_view, TRUE);
     data->on_screen = FALSE;
 }
 
 
+/***********************************************************************
+ *              sync_window_z_order
+ */
+static void sync_window_z_order(struct macdrv_win_data *data)
+{
+    if (data->cocoa_view)
+    {
+        HWND parent = GetAncestor(data->hwnd, GA_PARENT);
+        macdrv_view superview = macdrv_get_client_cocoa_view(parent);
+        macdrv_window window = NULL;
+        HWND prev;
+        HWND next = NULL;
+        macdrv_view prev_view = NULL;
+        macdrv_view next_view = NULL;
+
+        if (!superview)
+        {
+            window = macdrv_get_cocoa_window(parent, FALSE);
+            if (!window)
+                WARN("hwnd %p/%p parent %p has no Cocoa window or view in this process\n", data->hwnd, data->cocoa_view, parent);
+        }
+
+        /* find window that this one must be after */
+        prev = GetWindow(data->hwnd, GW_HWNDPREV);
+        while (prev && !(prev_view = macdrv_get_cocoa_view(prev)))
+            prev = GetWindow(prev, GW_HWNDPREV);
+        if (!prev_view)
+        {
+            /* find window that this one must be before */
+            next = GetWindow(data->hwnd, GW_HWNDNEXT);
+            while (next && !(next_view = macdrv_get_cocoa_view(next)))
+                next = GetWindow(next, GW_HWNDNEXT);
+        }
+
+        TRACE("win %p/%p below %p/%p above %p/%p\n",
+              data->hwnd, data->cocoa_view, prev, prev_view, next, next_view);
+
+        macdrv_set_view_superview(data->cocoa_view, superview, window, prev_view, next_view);
+    }
+    else if (data->on_screen)
+        show_window(data);
+}
+
+
 /***********************************************************************
  *              get_region_data
  *
@@ -855,27 +1017,64 @@ RGNDATA *get_region_data(HRGN hrgn, HDC hdc_lptodp)
 static void sync_window_position(struct macdrv_win_data *data, UINT swp_flags, const RECT *old_window_rect,
                                  const RECT *old_whole_rect)
 {
-    CGRect frame;
+    CGRect frame = cgrect_from_rect(data->whole_rect);
+    BOOL force_z_order = FALSE;
 
-    if (data->minimized) return;
+    if (data->cocoa_window)
+    {
+        if (data->minimized) return;
 
-    frame = cgrect_from_rect(data->whole_rect);
-    constrain_window_frame(&frame);
-    if (frame.size.width < 1 || frame.size.height < 1)
-        frame.size.width = frame.size.height = 1;
+        constrain_window_frame(&frame);
+        if (frame.size.width < 1 || frame.size.height < 1)
+            frame.size.width = frame.size.height = 1;
+
+        macdrv_set_cocoa_window_frame(data->cocoa_window, &frame);
+    }
+    else
+    {
+        BOOL were_equal = (data->cocoa_view == data->client_cocoa_view);
+        BOOL now_equal = !memcmp(&data->whole_rect, &data->client_rect, sizeof(data->whole_rect));
+
+        if (were_equal && !now_equal)
+        {
+            data->cocoa_view = macdrv_create_view(frame);
+            macdrv_set_view_hidden(data->cocoa_view, !data->on_screen);
+            macdrv_set_view_superview(data->client_cocoa_view, data->cocoa_view, NULL, NULL, NULL);
+            macdrv_set_view_hidden(data->client_cocoa_view, FALSE);
+            force_z_order = TRUE;
+        }
+        else if (!were_equal && now_equal)
+        {
+            macdrv_dispose_view(data->cocoa_view);
+            data->cocoa_view = data->client_cocoa_view;
+            macdrv_set_view_hidden(data->cocoa_view, !data->on_screen);
+            macdrv_set_view_frame(data->cocoa_view, frame);
+            force_z_order = TRUE;
+        }
+        else
+            macdrv_set_view_frame(data->cocoa_view, frame);
+    }
+
+    if (data->cocoa_view != data->client_cocoa_view)
+    {
+        RECT rect = data->client_rect;
+        OffsetRect(&rect, -data->whole_rect.left, -data->whole_rect.top);
+        macdrv_set_view_frame(data->client_cocoa_view, cgrect_from_rect(rect));
+        TRACE("win %p/%p client %s\n", data->hwnd, data->client_cocoa_view, wine_dbgstr_rect(&rect));
+    }
 
-    macdrv_set_cocoa_window_frame(data->cocoa_window, &frame);
     if (old_window_rect && old_whole_rect &&
         (IsRectEmpty(old_window_rect) != IsRectEmpty(&data->window_rect) ||
          old_window_rect->left - old_whole_rect->left != data->window_rect.left - data->whole_rect.left ||
          old_window_rect->top - old_whole_rect->top != data->window_rect.top - data->whole_rect.top))
         sync_window_region(data, (HRGN)1);
 
-    TRACE("win %p/%p whole_rect %s frame %s\n", data->hwnd, data->cocoa_window,
+    TRACE("win %p/%p whole_rect %s frame %s\n", data->hwnd,
+          data->cocoa_window ? (void*)data->cocoa_window : (void*)data->cocoa_view,
           wine_dbgstr_rect(&data->whole_rect), wine_dbgstr_cgrect(frame));
 
-    if (data->on_screen && (!(swp_flags & SWP_NOZORDER) || (swp_flags & SWP_SHOWWINDOW)))
-        show_window(data);
+    if (force_z_order || !(swp_flags & SWP_NOZORDER) || (swp_flags & SWP_SHOWWINDOW))
+        sync_window_z_order(data);
 }
 
 
@@ -1276,8 +1475,9 @@ void CDECL macdrv_DestroyWindow(HWND hwnd)
 
     if (hwnd == GetCapture()) macdrv_SetCapture(0, 0);
 
-    if (data->gl_view) macdrv_dispose_view(data->gl_view);
     destroy_cocoa_window(data);
+    destroy_cocoa_view(data);
+    if (data->client_cocoa_view) macdrv_dispose_view(data->client_cocoa_view);
 
     CFDictionaryRemoveValue(win_datas, hwnd);
     release_win_data(data);
@@ -1361,13 +1561,27 @@ void CDECL macdrv_SetParent(HWND hwnd, HWND parent, HWND old_parent)
         {
             /* destroy the old Mac window */
             destroy_cocoa_window(data);
+            create_cocoa_view(data);
+        }
+        else
+        {
+            struct macdrv_win_data *parent_data = get_win_data(parent);
+            macdrv_window cocoa_window = parent_data ? parent_data->cocoa_window : NULL;
+            macdrv_view superview = parent_data ? parent_data->client_cocoa_view : NULL;
+
+            if (!cocoa_window && !superview)
+                WARN("hwnd %p new parent %p has no Cocoa window or view in this process\n", hwnd, parent);
+
+            macdrv_set_view_superview(data->cocoa_view, superview, cocoa_window, NULL, NULL);
+            release_win_data(parent_data);
         }
     }
     else  /* new top level window */
+    {
+        destroy_cocoa_view(data);
         create_cocoa_window(data);
+    }
     release_win_data(data);
-
-    set_gl_view_parent(hwnd, parent);
 }
 
 
@@ -1848,9 +2062,9 @@ void CDECL macdrv_WindowPosChanged(HWND hwnd, HWND insert_after, UINT swp_flags,
         }
     }
 
-    sync_gl_view(data);
+    sync_gl_view(data, &old_whole_rect, &old_client_rect);
 
-    if (!data->cocoa_window) goto done;
+    if (!data->cocoa_window && !data->cocoa_view) goto done;
 
     if (data->on_screen)
     {
@@ -1860,22 +2074,28 @@ void CDECL macdrv_WindowPosChanged(HWND hwnd, HWND insert_after, UINT swp_flags,
 
     /* check if we are currently processing an event relevant to this window */
     if (!thread_data || !thread_data->current_event ||
-        thread_data->current_event->window != data->cocoa_window ||
+        !data->cocoa_window || thread_data->current_event->window != data->cocoa_window ||
         (thread_data->current_event->type != WINDOW_FRAME_CHANGED &&
          thread_data->current_event->type != WINDOW_DID_UNMINIMIZE))
     {
         sync_window_position(data, swp_flags, &old_window_rect, &old_whole_rect);
-        set_cocoa_window_properties(data);
+        if (data->cocoa_window)
+            set_cocoa_window_properties(data);
     }
 
     if (new_style & WS_VISIBLE)
     {
-        if (!data->on_screen || (swp_flags & (SWP_FRAMECHANGED|SWP_STATECHANGED)))
-            set_cocoa_window_properties(data);
+        if (data->cocoa_window)
+        {
+            if (!data->on_screen || (swp_flags & (SWP_FRAMECHANGED|SWP_STATECHANGED)))
+                set_cocoa_window_properties(data);
 
-        /* layered windows are not shown until their attributes are set */
-        if (!data->on_screen &&
-            (data->layered || !(GetWindowLongW( hwnd, GWL_EXSTYLE ) & WS_EX_LAYERED)))
+            /* layered windows are not shown until their attributes are set */
+            if (!data->on_screen &&
+                (data->layered || !(GetWindowLongW( hwnd, GWL_EXSTYLE ) & WS_EX_LAYERED)))
+                show_window(data);
+        }
+        else if (!data->on_screen)
             show_window(data);
     }
 
